// RuleMatrix.cpp : implementation file
//

#include "stdafx.h"
#include "Fuzzy.h"
#include "RuleMatrix.h"

#include "NewCellTypes/GridURLCell.h"
#include "NewCellTypes/GridCellCombo.h"
#include "NewCellTypes/GridCellCheck.h"
#include "NewCellTypes/GridCellNumeric.h"
#include "NewCellTypes/GridCellDateTime.h"

// CRuleMatrix dialog

IMPLEMENT_DYNAMIC(CRuleMatrix, CDialog)
CRuleMatrix::CRuleMatrix(CWnd* pParent /*=NULL*/)
	: CDialog(CRuleMatrix::IDD, pParent)
{
}

CRuleMatrix::~CRuleMatrix()
{
	// for each row, delete columns first
	for (int i = 0; i < m_nCols; i++)
		delete [] FuzzyRelations[i];
	delete [] FuzzyRelations;	// then delete rows
	FuzzyRelations = NULL;
}

void CRuleMatrix::DoDataExchange(CDataExchange* pDX)
{
	CDialog::DoDataExchange(pDX);
	DDX_GridControl(pDX, IDC_RULEMATRIX, m_Grid);
	DDX_Control(pDX, IDC_EDIT1, Var1);
	DDX_Control(pDX, IDC_EDIT2, Var2);
}

BEGIN_MESSAGE_MAP(CRuleMatrix, CDialog)
	ON_BN_CLICKED(IDOK, OnBnClickedOk)
	ON_BN_CLICKED(IDCANCEL, &CRuleMatrix::OnBnClickedCancel)
END_MESSAGE_MAP()

// CRuleMatrix message handlers

BOOL CRuleMatrix::OnInitDialog()
{
	CDialog::OnInitDialog();

	// Set up GridCtrl
	m_Grid.DeleteAllItems();
	m_Grid.EnableDragAndDrop(TRUE);
	m_Grid.GetDefaultCell(FALSE, FALSE)->SetBackClr(RGB(0xFF, 0xFF, 0xE0));

	m_Grid.EnableTitleTips(FALSE);

	m_Grid.SetFixedColumnSelection(TRUE);
    m_Grid.SetFixedRowSelection(TRUE);
    m_Grid.AutoSize();
	m_Grid.SetCompareFunction(CGridCtrl::pfnCellNumericCompare);

	m_nFixCols = 1;
	m_nFixRows = 1;

	m_Grid.SetGridLines(GVL_BOTH);
	m_Grid.SetAutoSizeStyle();

	BOOL bFound1, bFound2, bFound3;
	bFound1 = bFound2 = bFound3 = FALSE;

	int j=0;
	while ( j < nNumVars ) // 3 vars (2 inputs + 1 output)
	{
		if (VS[j].IsInput() && !bFound1)
		{	// 1st variable found
			m_nRows = VS[j].NumStates() + 1;
			bFound1 = TRUE;
			j++;
			continue;
		}
		if (VS[j].IsInput() && bFound1 && !bFound2)
		{	// 2nd variable found
			m_nCols = VS[j].NumStates() + 1;
			bFound2 = TRUE;
			j++;
			continue;
		}
		if (!VS[j].IsInput()) j++; // output
	}

	//	This 2D array will contain premise/consequent numbers
	//	to facilitate rule handling
	FuzzyRelations = new int*[m_nRows];
	for (int x=0; x<m_nRows; x++)
		FuzzyRelations[x] = new int[m_nCols];

	FuzzyRelations[0][0] = 0; // top left cell of rule array is empty

	TRY
	{
		m_Grid.SetRowCount(m_nRows);
		m_Grid.SetColumnCount(m_nCols);
		m_Grid.SetFixedRowCount(m_nFixRows);
		m_Grid.SetFixedColumnCount(m_nFixCols);
	}
	CATCH (CMemoryException, e)
	{
	    e->ReportError();
		return FALSE;
	}
    END_CATCH

	m_Grid.SetEditable(TRUE);
    
	CRect Rect;	// arrange columns width
	m_Grid.GetClientRect(&Rect);
	m_Grid.SetColumnWidth(0,80);
	for (int i=1;i<m_Grid.GetColumnCount();i++)
		m_Grid.SetColumnWidth(i,(Rect.Width()-80)/3);

	m_Grid.SetItemText(0,0,"AND");
	j = 0;
	bFound1 = bFound2 = FALSE;
	while (	j < nNumVars )
	{	// set up row and column headers
		if (VS[j].IsInput() && !bFound1)
		{
			Var1.SetWindowText( VS[j].Name().c_str() );
			bFound1 = TRUE;
			CString str, strDisp;
			for (int k = 0; k < VS[j].NumStates(); k++)
			{
				FuzzyRelations[0][k+1] = VS[j].FuzzySet()->FuzzyStateElt(k).PremiseNumber();
				//str.Format("%i ", FuzzyRelations[0][k+1]);
				//strDisp.Format("%s (%s)", 
				//	VS[j].FuzzySet()->FuzzyStateElt(k).GetName().c_str(), 
				//	str);
				CString s = VS[j].FuzzySet()->FuzzyStateElt(k).GetName().c_str();
				strDisp.Format("%s (%i)", 
					s, 
					FuzzyRelations[0][k+1]);
				m_Grid.SetItemText(	0, k+1, strDisp);
			}
			j++;
			continue;	// go to next variable
		}
		if (VS[j].IsInput() && bFound1 && !bFound2)
		{
			Var2.SetWindowText( VS[j].Name().c_str() );
			bFound2 = TRUE;
			CString str, strDisp;
			for (int k = 0; k < VS[j].NumStates(); k++)
			{
				FuzzyRelations[k+1][0] = VS[j].FuzzySet()->FuzzyStateElt(k).PremiseNumber();
				//str.Format("%i ", FuzzyRelations[k+1][0]);
				//strDisp.Format("%s (%s)", 
				//	VS[j].FuzzySet()->FuzzyStateElt(k).GetName().c_str(), 
				//	str);
				strDisp.Format("%s (%i)", 
					VS[j].FuzzySet()->FuzzyStateElt(k).GetName().c_str(), 
					FuzzyRelations[k+1][0]);
				m_Grid.SetItemText( k+1, 0, strDisp);
			}
			j++;
			continue;	// go to next variable
		}
		// end setup row & column headers
		if (!VS[j].IsInput())	// output
		{
		    CStringArray OutputStates;
			for (int k=0; k<VS[j].NumStates(); k++)
			{
				CString str, strDisp;
				//str.Format(" (%i)", VS[j].FuzzySet()->FuzzyStateElt(k).PremiseNumber());
				//strDisp.Format("%s %s",
				//	VS[j].FuzzySet()->FuzzyStateElt(k).GetName().c_str(),
				//	str);
				strDisp.Format("%s (%i)",
					VS[j].FuzzySet()->FuzzyStateElt(k).GetName().c_str(),
					VS[j].FuzzySet()->FuzzyStateElt(k).PremiseNumber());
				OutputStates.Add(strDisp);
			}
			
			for (int m=1; m<m_Grid.GetRowCount(); m++)
				for (int n=1; n<m_Grid.GetColumnCount(); n++)
				{	// set combocells
                    m_Grid.SetCellType(m,n, RUNTIME_CLASS(CGridCellCombo));
				    CGridCellCombo *pCell1 = (CGridCellCombo*) m_Grid.GetCell(m,n);
					pCell1->SetOptions(OutputStates);
					pCell1->SetStyle(CBS_DROPDOWN); //CBS_DROPDOWN, CBS_DROPDOWNLIST, CBS_SIMPLE
				}

			int X,Y;
			X=Y=0;
			for (int r=0; r<nNumRules; r++)
			{	// load states from rule set
				int np = RS[r].NumPremises(); // 2, of course
				for (int p = 0; p < np; p++)
				{
					for (int m=1; m<m_Grid.GetRowCount(); m++)
						if (RS[r].Premise(p) == FuzzyRelations[m][0])
						{
							X = m;
							break;
						}
					for (int n=1; n<m_Grid.GetColumnCount(); n++)
						if (RS[r].Premise(p) == FuzzyRelations[0][n])
						{
							Y = n;
							break;
						}
				}
				CFLMembershipFunction * pMF = VS[j].FuzzySet()->FuzzyStateArray();
				BOOL found = FALSE;
				for (int i=0; i<VS[j].NumStates(); i++)
					if (pMF[i].PremiseNumber() == RS[r].RuleConsequent())
					{
						CString str = pMF[i].GetName().c_str();
						CString str2;
						str2.Format(" (%i)", pMF[i].PremiseNumber());
						m_Grid.SetItemText(X,Y,str + str2);
						found = TRUE;
						break;
					}
				if (!found || X==0 || Y==0 ) m_Grid.SetItemText(X,Y,"Error!");
			}
			j++;
			bFound3 = TRUE;
		}
	}

	if (!bFound1) AfxMessageBox("Input1 not found!!!", MB_ICONEXCLAMATION | MB_OK);
	if (!bFound2) AfxMessageBox("Input2 not found!!!", MB_ICONEXCLAMATION | MB_OK);
	if (!bFound3) AfxMessageBox("Output not found!!!", MB_ICONEXCLAMATION | MB_OK);

	return TRUE;
}

/*
void CRuleMatrix::OnCbnSelchangeCombo1()
{
	int n = m_FStatesNames.GetCurSel();
	for (int i=1; i<=m_nRows; i++)
		if (m_nRows == 5)
			m_Grid.SetItemText(i,1,strStates5[n][i-1]);
		else
			m_Grid.SetItemText(i,1,strStates3[n][i-1]);
	m_Grid.Refresh();
}
*/

void CRuleMatrix::OnBnClickedOk()
{
	int j = 0;
	while (j<nNumVars)
	{	
		if (!VS[j].IsInput())	// output
		{
			int X,Y;
			X=Y=0;
			for (int r=0; r<nNumRules; r++)
			{	// load states from rule set
				int np = RS[r].NumPremises(); // 2, of course
				for (int p=0; p<np; p++)
				{
					for (int m=1; m<m_Grid.GetRowCount(); m++)
						if (RS[r].Premise(p) == FuzzyRelations[m][0])
						{
							X = m;
							break;
						}
					for (int n=1; n<m_Grid.GetColumnCount(); n++)
						if (RS[r].Premise(p) == FuzzyRelations[0][n])
						{
							Y = n;
							break;
						}
				}
				// found it, now update 
				CFLMembershipFunction * pMF = VS[j].FuzzySet()->FuzzyStateArray();
				for (int i=0; i<VS[j].NumStates(); i++)
				{
					CString strState = m_Grid.GetItemText(X,Y);
					if (pMF[i].GetName().c_str() == strState.GetString())
						RS[r].SetConsequent(pMF[i].PremiseNumber());
				}
			}
			
		}
		j++;
	}


	OnOK();
}

void CRuleMatrix::OnBnClickedCancel()
{
	// TODO: Add your control notification handler code here
	OnCancel();
}
