// FuzzyDlg.cpp : implementation file
//

//	This is an implementation of a FL Controller as per the 
//	SeattleRobotics.org tutorial
//	http://www.seattlerobotics.org/encoder/mar98/fuz/flindex.htm
//
//	This source code is not to be distributed without prior consent of its author 
//
//	BV 2003 - 2008

#include "stdafx.h"
#include "FLC_DLL.h"
#include "FuzzyDlg.h"
#include "FuzzySetDesign.h"
#include "FuzzyVarDesign.h"
#include "RuleMatrix.h"
#include "FS_shape.h"

#ifdef _DEBUG
#define new DEBUG_NEW
#endif

// CAboutDlg dialog used for App About
class CAboutDlg : public CDialog
{
public:
	CAboutDlg();

// Dialog Data
	enum { IDD = IDD_ABOUTBOX };

	protected:
	virtual void DoDataExchange(CDataExchange* pDX);    // DDX/DDV support

// Implementation
protected:
	DECLARE_MESSAGE_MAP()
};

CAboutDlg::CAboutDlg() : CDialog(CAboutDlg::IDD)
{
}

void CAboutDlg::DoDataExchange(CDataExchange* pDX)
{
	CDialog::DoDataExchange(pDX);
}

BEGIN_MESSAGE_MAP(CAboutDlg, CDialog)
END_MESSAGE_MAP()


// CFuzzyDlg dialog

CFuzzyDlg::CFuzzyDlg(CWnd* pParent /*=NULL*/)
	: CDialog(CFuzzyDlg::IDD, pParent)
	, Input1(0)
	, Input2(0)
	, Output(0)
	, m_strFLCname(_T(""))
	, m_strRuleName(NULL)
	, m_strFuzzySetName(NULL)
	, m_strVariableName(NULL)
	, m_nDefaultFuzzySetModel(-1)
	, m_nDefaultRuleSetModel(-1)
	, m_nFLC_Shape(-1)
{
	m_hIcon = AfxGetApp()->LoadIcon(IDI_ICON);
	FLC = new CFLC;		// instantiate a new FL 'empty' controller
	m_strFLCname.Format(_T("FLC_%i"),FLC->FLC_Number()); 
}

CFuzzyDlg::~CFuzzyDlg()
{
	if (FLC) 
	{
		FLC->Reset();
		delete FLC;
	}
	if (m_strRuleName)
		delete [] m_strRuleName;
	if (m_strFuzzySetName)			// display name for fuzzy sets
		delete [] m_strFuzzySetName;
	if (m_strVariableName)			// display name for variables
		delete [] m_strVariableName;
}

void CFuzzyDlg::DoDataExchange(CDataExchange* pDX)
{
	CDialog::DoDataExchange(pDX);
	DDX_Text(pDX,IDC_EDIT1,Input1);
	DDX_Text(pDX,IDC_EDIT2,Input2);
	DDX_Text(pDX,IDC_EDIT3,Output);
	DDX_Control(pDX, IDC_LIST1, lstPredicate);
	DDX_Control(pDX, IDC_LIST3, lstOutputStrength);
	DDX_Control(pDX, IDC_LIST4, lstEvalPredicates);
	DDX_Control(pDX, IDC_EDIT3, FL_output);
	DDX_Control(pDX, IDC_LIST2, m_Info);
	DDX_Control(pDX, IDC_COMBO1, cbbFS);
	DDX_Control(pDX, IDC_COMBO2, cbbRS);
	DDX_Control(pDX, IDCANCEL, button_33or55);
	DDX_Control(pDX, IDC_CHECK1, btnVerbose);
	DDX_Control(pDX, IDC_CHECK2, m_chkInvVar1);
	DDX_Control(pDX, IDC_CHECK3, m_chkInvVar2);
	DDX_Control(pDX, IDC_CHECK5, m_chkInvOut);
	DDX_Control(pDX, IDC_INVRULES, m_chkInvRules);
}

BEGIN_MESSAGE_MAP(CFuzzyDlg, CDialog)
	ON_WM_SYSCOMMAND()
	ON_WM_PAINT()
	ON_WM_QUERYDRAGICON()
	//}}AFX_MSG_MAP
	ON_BN_CLICKED(IDOK, sFLC_Calculate)
	ON_BN_CLICKED(IDCANCEL, sFLC_Close)
	ON_BN_CLICKED(IDC_BUTTON1, ModifyFS)
	ON_BN_CLICKED(IDC_BUTTON2, ModifyVS)
	ON_BN_CLICKED(IDC_BUTTON3, ModifyRS)
	ON_BN_CLICKED(IDC_BUTTON4, ClearInfoList)
	ON_BN_CLICKED(IDC_BUTTON5, LoadDefaultFS)
	ON_BN_CLICKED(IDC_BUTTON6, LoadDefaultVS)
	ON_BN_CLICKED(IDC_BUTTON7, LoadDefaultRS)
	ON_BN_CLICKED(IDC_BUTTON8, DeleteFS)
	ON_BN_CLICKED(IDC_BUTTON9, DeleteVS)
	ON_BN_CLICKED(IDC_BUTTON10, DeleteRS)
	ON_BN_CLICKED(IDC_BUTTON11, ClearFLCDisplayList)
	ON_WM_DESTROY()
	ON_BN_CLICKED(IDC_RADIO1, &CFuzzyDlg::Click_3x3_FS_model)
	ON_BN_CLICKED(IDC_RADIO2, &CFuzzyDlg::Click_5x5_FS_model)
	ON_CBN_SELCHANGE(IDC_COMBO1, &CFuzzyDlg::Select_FSmodel)
	ON_CBN_SELCHANGE(IDC_COMBO2, &CFuzzyDlg::Select_RSmodel)
	ON_BN_CLICKED(IDC_BUTTON12, &CFuzzyDlg::OnBnClickedShape)
	ON_BN_CLICKED(IDC_WEB, &CFuzzyDlg::OnBnClickedWeb)
END_MESSAGE_MAP()

// CFuzzyDlg message handlers

BOOL CFuzzyDlg::OnInitDialog()
{
	CDialog::OnInitDialog();
	m_Info.ResetContent();
	//
	//	A new sFLC has been created in the constructor
	//  A default name is given, and has been slight modified 
	//	with a FLC_ prefix in the constructor as well
	
	AddInfoLine("New sFLC object (no:" + m_strFLCname + ") has been instantiated");
	AddInfoLine("This newly created sFLC has no FS, no VS, no RS yet");
	AddInfoLine("");
	AddInfoLine("So please first select FS and RS format (3x3 or 5x5 in this sample progam)");
	AddInfoLine("And a default model for fuzzy sets and rule set (click on shape to display FS behaviour)");
	AddInfoLine("Then define, construct and run your custom sFLC");
	AddInfoLine("");
	AddInfoLine("You can notice from v3.06 that the FLC can be 'twisted' further");
	AddInfoLine("by inverting variables or rules, for better versatility");
	AddInfoLine("without adding new models 'artificially'.");
	AddInfoLine("for more info, click on 'www' button above to reach web page.");


	// IDM_ABOUTBOX must be in the system command range.
	ASSERT((IDM_ABOUTBOX & 0xFFF0) == IDM_ABOUTBOX);
	ASSERT(IDM_ABOUTBOX < 0xF000);

	CMenu* pSysMenu = GetSystemMenu(FALSE);
	if (pSysMenu != NULL)
	{
		CString strAboutMenu;
		strAboutMenu.LoadString(IDS_ABOUTBOX);
		if (!strAboutMenu.IsEmpty())
		{
			pSysMenu->AppendMenu(MF_SEPARATOR);
			pSysMenu->AppendMenu(MF_STRING, IDM_ABOUTBOX, strAboutMenu);
		}
	}

	SetIcon(m_hIcon, TRUE);			// Set big icon
	SetIcon(m_hIcon, FALSE);		// Set small icon

	btnVerbose.SetCheck(TRUE);

	//	This simple sFLC example uses 2 inputs, 
	//	Input1 and Input2 here set to 0
	Input1 = 0.0;
	Input2 = 0.0;
	return TRUE;
}

void CFuzzyDlg::OnSysCommand(UINT nID, LPARAM lParam)
{
	if ((nID & 0xFFF0) == IDM_ABOUTBOX)
	{
		CAboutDlg dlgAbout;
		dlgAbout.DoModal();
	}
	else
		CDialog::OnSysCommand(nID, lParam);
}

void CFuzzyDlg::OnPaint() 
{
	if (IsIconic())
	{
		CPaintDC dc(this); // device context for painting
		SendMessage(WM_ICONERASEBKGND, reinterpret_cast<WPARAM>(dc.GetSafeHdc()), 0);
		// Center icon in client rectangle
		int cxIcon = GetSystemMetrics(SM_CXICON);
		int cyIcon = GetSystemMetrics(SM_CYICON);
		CRect rect;
		GetClientRect(&rect);
		int x = (rect.Width() - cxIcon + 1) / 2;
		int y = (rect.Height() - cyIcon + 1) / 2;
		// Draw the icon
		dc.DrawIcon(x, y, m_hIcon);
	}
	else
		CDialog::OnPaint();
}

// The system calls this function to obtain the cursor to display while the user drags
//  the minimized window.
HCURSOR CFuzzyDlg::OnQueryDragIcon()
{
	return static_cast<HCURSOR>(m_hIcon);
}

void CFuzzyDlg::sFLC_Calculate()
{	
	// The sFLC is supposed to have been set up
	// We can now proceed with some calculation

	m_Info.ResetContent();

	if (FLC->GetShape() < 0)
	{	// Shape defined by clicking 3x3 or 5x5
		AddInfoLine("sFLC is not defined. Can't continue...");
		return;
	}

	if (!FLC->boolFS())
	{
		AddInfoLine("No Fuzzy Sets. Can't continue...");
		return;
	}

	if (!FLC->boolVS())
	{
		AddInfoLine("No Variable Set. Can't continue...");
		return;
	}

	if (!FLC->boolRS())
	{
		AddInfoLine("No Rule Set. Can't continue...");
		return;
	}

	lstPredicate.ResetContent();
	lstEvalPredicates.ResetContent();
	lstOutputStrength.ResetContent();
	UpdateData(TRUE);

	if (m_chkInvVar1.GetCheck() != 0) 
	{
		Input1 = -Input1;
		AddInfoLine("Input1 is to be inverted");
	}
	if (m_chkInvVar2.GetCheck() != 0) 
	{
		Input2 = -Input2;
		AddInfoLine("Input2 is to be inverted");
	}


	double ArrInput[2];
	ArrInput[0] = Input1;	// as set by DDX
	ArrInput[1] = Input2;	// as set by DDX

	//	This sFLC client uses one call to calculate the fuzzy output
	//	The format is self explanatory
	//	One could of course devise other ways to pass data etc.
	//	FuzzyController set the sFLC output
	double ret = FuzzyController(ArrInput);

	// Output retrieved then displayed
	int i = (0 == m_chkInvOut.GetCheck()) ? 1 : -1;
	Output = i * FLC->Output();

	CString str;
	str.Format("%.3f",Output);
	FL_output.SetWindowText(str);
}

double CFuzzyDlg::FuzzyController(double *VarValue)
{
//  Besides the usual checks, this routine does:
//  Calculate premises for each input
//
	CString str;
	double OutputValue = 0.0;

	// In order to be operational, FLC will need FuzzySets, FuzzyVars, and a Rule Matrix
	if (!FLC->boolFS())
	{
		AddInfoLine("Fuzzy Sets not loaded. Can't continue...");
		return 0.00;
	}

	if (!FLC->boolVS())
	{
		AddInfoLine("Variable Set not loaded. Can't continue...");
		return 0.00;
	}

	if (!FLC->boolRS())
	{
		AddInfoLine("Rule Set not loaded. Can't continue...");
		return 0.00;
	}

	// STEP 1
	FLC->EvaluateAllPremises(VarValue);
	// END STEP 1
	// DISPLAY CODE
	CFLVariable * pVars = FLC->VarArray();
	int nv = FLC->GetNumVars();
	CString strVal;
	for (int i = 0; i < nv ; i++)
	{	// we calc premises for each variable
		if (pVars[i].IsInput())
		{
			CFuzzySet * vFS = pVars[i].FuzzySet();
			CFLMembershipFunction * pMF = vFS->FuzzyStateArray();
			for (int j = 0; j < vFS->NumStates(); j++)
			{
				CString name = pMF[j].GetName().c_str();
				double dom = pMF[j].GetDOM();
				int nPN = pMF[j].PremiseNumber();
				strVal.Format("V%i (p%i) is %.0f%% %s",
				i, nPN,
				100*dom,	// to display %
				name);
				lstPredicate.AddString(strVal);
			}
		}
	}
	AddInfoLine("All fuzzy states are now evaluated (left list above)");
	// END DISPLAY CODE

	if (!FLC->boolRS())
	{
		AddInfoLine("Rule Set not loaded. Can't continue...");
		return 0.00;
	}

	// STEP 2
	FLC->InferenceRules();
	// END STEP 2

	// DISPLAY CODE
	// Rules are composed of a Predicate block followed by a Consequent Block
	// i.e. If Predicate then Consequent
	// with Predicate = Premise1 AND/OR Premise2 ...

	CFLRule * pFLR = FLC->RuleArray();	// get a pointer to the rule set

	int NumRules = FLC->GetRuleSetSize();

	//	Now, since premises have already been evaluated previously
	//	all is left to do really is inferencing the rules using input values

	int np = FLC->RuleArray()->NumPremises();	// 2
	
	// The rules will now be evaluated.
	// We build an array of Rule Estimates 
	double *PremiseEstimate = new double[np];
	for (int i = 0 ; i < NumRules ; i++)
	{
		strVal.Format(_T("Rule:%i between premises"),i+1);
		// An array of doubles will contain each premise evaluation
		// Then the overall Predicate is calculated
		for (int n = 0; n < np ; n++)
		// Premise numbers are unique
		// One just needs to find which variable it can be associated with
		{
			CString str;
			int *pP = pFLR[i].pPremise();	// get a pointer to premises in current rule
			str.Format(" %i", pP[n] );
			PremiseEstimate[n] = 0.0;
			for (int v = 0 ; v < nv ; v++)		// NumVariables = 3 in this example (2 inputs)
			{
				CFLMembershipFunction * pMF = pVars[v].FuzzySet()->FuzzyStateArray();
				if (pVars[v].IsInput())	// Only rule antecedents need to be evaluated
					for (int k = 0 ; k < pVars[v].NumStates() ; k++)	// Num FuzzyStates for each variable = 3
						if (pP[n] == pMF[k].PremiseNumber())
							PremiseEstimate[n] = pMF[k].GetDOM();
			}
			if (n < np - 1) str+= " and";
			strVal+=str;
		}
		lstEvalPredicates.AddString(strVal);

		double value = 0.0;
	
		CString strDisplay;	
		// Now apply "fuzzy" booleans to Premises to evaluate the Predicate
		for (int p = 0 ; p < np ; p++)
		{
			int o = pFLR[i].Operation(p);
			switch (o)
			{
				case AND:
					value += fbAnd(PremiseEstimate[p],PremiseEstimate[p+1]);
					strDisplay.Format(_T("AND(%.3f,%.3f)=%.3f"),PremiseEstimate[p],PremiseEstimate[p+1],value);
					break;
				case OR:
					value += fbOr(PremiseEstimate[p],PremiseEstimate[p+1]);
					strDisplay.Format(_T("OR(%.3f,%.3f)=%.3f"),PremiseEstimate[p],PremiseEstimate[p+1],value);
					break;
				case NOT:
					value += fbNot(PremiseEstimate[p]);
					strDisplay.Format(_T("NOT(%.3f)=%.3f"),PremiseEstimate[p],value);
					break;
				case NOTHING:
					// Do nothing
					// End predicate evaluation
				default:
				;
			}
		}
		lstEvalPredicates.AddString(strDisplay);
	}
	
	delete [] PremiseEstimate;	// no longer needed
	PremiseEstimate = NULL;

	AddInfoLine("Predicates are calculated for each rule (middle pane)");
	
	// END DISPLAY CODE

	// STEP 3
	FLC->GenerateFuzzyOutputs();
	// END STEP 3
	AddInfoLine("Fuzzy output states are calculated (right pane)");

	// STEP 4
	FLC->Defuzzify(0);
	// END STEP 4
	AddInfoLine("Defuzzification done (Centroid) into single output");

	// DISPLAY CODE
	int OutVarNo = FLC->GetNumVars() - 1;
	CFuzzySet *vFS = pVars[OutVarNo].FuzzySet();
	int NumOutputStates = vFS->NumStates();

	for (int i = 0; i < NumOutputStates; i++)
	{
		// Display
		strVal.Format(_T("State %i: %.0f%% %s"),
			i,
			100*FLC->OutStates()[i],
			vFS->FuzzyStateArray()[i].GetName().c_str()	);
		lstOutputStrength.AddString(strVal);
	}
	// END DISPLAY CODE

	return 0.0;
}

void CFuzzyDlg::sFLC_Close()
{	// clean up and go
	FLC->Reset();
	OnCancel();
}

void CFuzzyDlg::LoadDefaultFS()
{
	CString str;
	const CString spc = " - ";

	//  Loading a default FS is simple but is also essential to building a good FLC
	//  One needs to check up on a few things
	//  but the core of this routine is only 3 calls
	//	InitFuzzySets()
	//  Get a pointer to the FLC's FuzzySets
	//	Then InitFuzzySet() and DefineFuzzySet

	//  Which shape is the FLC? If not defined, just exit
	if (FLC->GetShape() < 0)
	{	// Shape is defined when selected FLC shape (3x3, 5x5)
		AddInfoLine("FLC Setup is not defined (3x3, 5x5 etc)");
		return;
	}

	//	Is there already a FS? (One could automatically delete it)
	if (FLC->boolFS())
	{
		AddInfoLine("Fuzzy Sets are already loaded");
		return;
	}
	AddInfoLine("3 default Fuzzy Sets are now being created:");
	AddInfoLine("There could be many more FS to choose from, and to attach to each variable");
	AddInfoLine("In this example, we assume a single fixed FS per variable");
	
	//// INIT FUZZY SETS
	// We now create 3 empty Fuzzy Sets for this FLC
	// (Caution: InitFuzzySets at FLC level, not to be confused with InitFuzzySet
	// We can create as many as we want.
	int ret = FLC->InitFuzzySets(3);
	if (ret < 0) return;	// if something went wrong
	//// END INIT

	// Since we used the default constructor, we need to initialise the Fuzzy Sets
	// First we get a pointer to (the array of) Fuzzy Sets we've just created
	// through a call to FSArray() method
	CFuzzySet* pFSet = FLC->FSArray();

	//  We know how many FS we have but just in case
	int nNumFuzzySets = FLC->GetNumFuzzySets();
	// Name Fuzzy Sets (optional)
	m_strFuzzySetName = new CString[nNumFuzzySets];
	if (btnVerbose.GetCheck())
	{
		for (int i = 0; i < nNumFuzzySets; i++)
		{
			m_strFuzzySetName[i] = pFSet[i].GetFSName().c_str();
			str.Format("FuzzySet no%i : %s",i,m_strFuzzySetName[i]);
			AddInfoLine(str);
		}
	}

	// each FS (here 3 or 5 in total) is now initialised with
	// i.e. 3 fuzzy states : Neg,Null,Pos (Wording = 1)

	int nNumFuzzyStates;
	switch (FLC->GetShape())	// given by radio buttons (3x3 or 5x5)
	{						// other models could easily be devised.
	default:				// for the time being, only 0 or 1
	case 0:
		nNumFuzzyStates = 3;	// to be used in a 3x3 rule set
		break;
	case 1:
		nNumFuzzyStates = 5;	// to be used in a 5x5 rule set
		break;
	case 2:						// not implemented
		nNumFuzzyStates = 7;	// to be used in a 7x7 rule set
		break;
	}

	if (btnVerbose.GetCheck()) 
		AddInfoLine("Each Fuzzy State is given a Unique Tag or Label:");
	for (int i = 0; i < nNumFuzzySets; i++)
	{
		//// DEFINE EACH FUZZY SET
		//pFSet[i].InitFuzzySet(nNumFuzzyStates,0);	
		pFSet[i].InitFuzzySet(nNumFuzzyStates,i);	
		pFSet[i].DefineFuzzySet(m_nDefaultFuzzySetModel); // before any possible change
		//// END DEFINE
		//	The rest is left for display purposes
		CFLMembershipFunction *MF = pFSet[i].FuzzyStateArray();
		switch (FLC->GetShape())
		{
		default:
		case 0:
			str.Format("FS%i Shape (Label): %s (%s) /%s (%s) / %s (%s)",
				i,
				MF[0].GetShapeName().c_str(), MF[0].GetName().c_str(),
				MF[1].GetShapeName().c_str(), MF[1].GetName().c_str(),
				MF[2].GetShapeName().c_str(), MF[2].GetName().c_str());
			break;
		case 1:
			str.Format("FS%i Shape (Label): %s (%s) /%s (%s) / %s (%s) /%s (%s) / %s (%s)",
				i,
				MF[0].GetShapeName().c_str(), MF[0].GetName().c_str(),
				MF[1].GetShapeName().c_str(), MF[1].GetName().c_str(),
				MF[2].GetShapeName().c_str(), MF[2].GetName().c_str(),
				MF[3].GetShapeName().c_str(), MF[3].GetName().c_str(),
				MF[4].GetShapeName().c_str(), MF[4].GetName().c_str());		
			break;
		case 2:
			str.Format("FS%i Shape (Label): %s (%s) /%s (%s) / %s (%s) /%s (%s) / %s (%s) / %s (%s) / %s (%s)",
				i,
				MF[0].GetShapeName().c_str(), MF[0].GetName().c_str(),
				MF[1].GetShapeName().c_str(), MF[1].GetName().c_str(),
				MF[2].GetShapeName().c_str(), MF[2].GetName().c_str(),
				MF[3].GetShapeName().c_str(), MF[3].GetName().c_str(),
				MF[4].GetShapeName().c_str(), MF[4].GetName().c_str(),
				MF[5].GetShapeName().c_str(), MF[5].GetName().c_str(),
				MF[6].GetShapeName().c_str(), MF[6].GetName().c_str());		
			break;
		}
		if (btnVerbose.GetCheck()) AddInfoLine(str);
	}

	if (FLC->boolFS()) 
		AddInfoLine("All default Fuzzy Sets are loaded into current FLC");
	else
		AddInfoLine("Default Fuzzy Sets not loaded into current FLC !!!");
}

void CFuzzyDlg::LoadDefaultVS() // Create set of Vars and Load Var Defaults
{
	//	The logic is simple:
	//	One cannot create an VS is there is one already
	//	One needs a FS to attach to a VS

	if (FLC->boolVS())
	{
		AddInfoLine("Variable Set already loaded");
		return;
	}

	if (!FLC->boolFS())
	{
		AddInfoLine("No Fuzzy Set. Can't continue...");
		return;
	}

	AddInfoLine("3 Variables are now being created (2 inputs + 1 output)");
	// First, a set of 3 variables is created in the FLC
	FLC->InitVarSet(3);	// (default constructor for variables)
	//	The pool of variables is private to FLC
	//	But one use a public const pointer to access it
	CFLVariable* pVars = FLC->VarArray();
	CFuzzySet* pFSet = FLC->FSArray();

	// 2 inputs, 1 output: each is associated with a specific fuzzy set as defined above
	// (default fuzzy sets are actually identical
	// New Fuzzy Sets are created, and existing FS are copied into them

/*	// One possible syntax:
	pVars[0].LoadFuzzySet(pFSet[0],TRUE);	// This is Input1 in the FL Controller
	pVars[1].LoadFuzzySet(pFSet[1],TRUE);	// This is Input2
	pVars[2].LoadFuzzySet(pFSet[2],FALSE);	// This is the rule output
*/	// Or in a loop:
	int nv = FLC->GetNumVars();	// Num variables
	for (int i = 0; i < nv; i++)
	{
		CString str, strIO;
		// Assuming here the last variable is the output
		BOOL bInputOutput = (i == nv - 1) ? FALSE : TRUE;
		strIO = bInputOutput ? "INPUT" : "OUTPUT";
		// CODE
		pVars[i].LoadFuzzySet(pFSet[i],bInputOutput);
		// END CODE
		CFuzzySet *vFS = pVars[i].FuzzySet();
		str.Format("%i) Fuzzy Set %s is now attached to Variable %s (%s)", i,
			pFSet[i].GetFSName().c_str(),
			pVars[i].Name().c_str(),
			strIO);
		AddInfoLine(str);
		if (btnVerbose.GetCheck())
		{
			for (int k = 0; k < pVars[i].NumStates(); k++)
			{
				string s = vFS->FuzzyStateArray()[k].GetName();
				str.Format("Var%i State%i: %s", i+1,k+1,s.c_str());
				AddInfoLine(str);
			}
		}
	}

// The 2 formats below are identical
/*	str.Format("Premises\nMF0 %i %i %i\nMF1 %i %i %i\nMF2 %i %i %i",
		pVars[0].FuzzySet()->FuzzyStateArray()[0].PremiseNumber(),
		pVars[0].FuzzySet()->FuzzyStateArray()[1].PremiseNumber(),
		pVars[0].FuzzySet()->FuzzyStateArray()[2].PremiseNumber(),
		pVars[1].FuzzySet()->FuzzyStateArray()[0].PremiseNumber(),
		pVars[1].FuzzySet()->FuzzyStateArray()[1].PremiseNumber(),
		pVars[1].FuzzySet()->FuzzyStateArray()[2].PremiseNumber(),
		pVars[2].FuzzySet()->FuzzyStateArray()[0].PremiseNumber(),
		pVars[2].FuzzySet()->FuzzyStateArray()[1].PremiseNumber(),
		pVars[2].FuzzySet()->FuzzyStateArray()[2].PremiseNumber()
		);

	str.Format("Premises\nMF0 %i %i %i\nMF1 %i %i %i\nMF2 %i %i %i",
		pVars[0].FuzzySet()->FuzzyStateElt(0).PremiseNumber(),
		pVars[0].FuzzySet()->FuzzyStateElt(1).PremiseNumber(),
		pVars[0].FuzzySet()->FuzzyStateElt(2).PremiseNumber(),
		pVars[1].FuzzySet()->FuzzyStateElt(0).PremiseNumber(),
		pVars[1].FuzzySet()->FuzzyStateElt(1).PremiseNumber(),
		pVars[1].FuzzySet()->FuzzyStateElt(2).PremiseNumber(),
		pVars[2].FuzzySet()->FuzzyStateElt(0).PremiseNumber(),
		pVars[2].FuzzySet()->FuzzyStateElt(1).PremiseNumber(),
		pVars[2].FuzzySet()->FuzzyStateElt(2).PremiseNumber()
		);
*/

	// Since the 'variable <-> FS' association is now done, there is no need
	// in this simple FL controller to establish a separate Premise object.
	// They are already all there.

	// Premises are numbered:
	// Each fuzzy set contains 3 states here, so:
	// pVars[0]	: premises 1 to 3 (from negative to positive)
	// pVars[1]	: premises 4 to 6
	// pVars[2]	: premises 7 to 9 (Maybe not rigorous to call them premises for a Consequent)
}

void CFuzzyDlg::LoadDefaultRS() // Load default Rule Set
{
	//	This call not only determines the number of rules needed
	//  for a complete Rule Set, but also checks all variables have 
	//  an attached Fuzzy Set
	int ret = FLC->DetermineRuleSetSize();
	if (ret  == -1)	// We first check a variable set has been defined
	{
		AddInfoLine("No Variable Set. Can't continue.");
		Beep(250,100);
		return;
	}

	if (ret < 0) // We then check that a Fuzzy Set has been attached to each Variable
	{	// Since ret = -1 is the 1st error code. Var Number is -(i+2)
		CString str;
		string s = FLC->VarArray()[-ret-2].Name();
		str = s.c_str();
		str.Append(" has no Fuzzy Set!");
		AddInfoLine(str);
		return;
	}
	else
	{
		CString strNR;
		strNR.Format("Num rules needed: %i", ret);
		if (btnVerbose.GetCheck())
			AddInfoLine(strNR);
	}

	//  Check a Rule Set has not been loaded already
	if (FLC->boolRS())
	{
		AddInfoLine("Rule Set already loaded");
		Beep(500,50);
		return;
	}



	int irs = FLC->InitRuleSet();	// generate an array of rules (RuleSetSize can be given or not)
	if (irs < 0)
	{
		AddInfoLine("Rule Set Init failed");
		Beep(500,50);
		return;
	}
	CFLRule *pFLR = FLC->RuleArray();	// get a pointer to the rule set

	// We retrieve NumRules (for test purposes only)
	int NumRules = FLC->GetRuleSetSize();

	// In this example, rule predicates are all made of 2 premises
	// i.e. If a AND b THEN
	// We now allocate some space to define the rules (here "A and B implies C"
	// i.e. made of 2 premises, 1 for each input variable.
	// This call is required before loading a Rule Set in order to
	// then just 'fill spaces' i.e. premises and operators
	ret = FLC->AllocateRuleSpace();
	if (ret < 0) 
	{
		AddInfoLine("Rule Space could not be allocated");
		return;
	}

	// Invert rules if so desired
	FLC->SetInvertRS( m_chkInvRules.GetCheck() != 0);

	// Now the rule set can be loaded in the FLC
	int RS_ret = -1;
	if (FLC->GetShape() == 0)
		RS_ret = FLC->LoadDefault33RuleSet(pFLR, NumRules,m_nDefaultRuleSetModel);
	else
		RS_ret = FLC->LoadDefault55RuleSet(pFLR, NumRules,m_nDefaultRuleSetModel);

	if (RS_ret != 0)
		AddInfoLine("Rule Set could not be loaded");
	else
		AddInfoLine("Default Rule Set is now loaded");
	CString str;
	//string s =
	//str = s.c_str();
	//str.Append(" has no Fuzzy Set!");
	//AddInfoLine(str);
}

void CFuzzyDlg::ModifyFS()
{
	//		We here modify previously created fuzzy sets (LoadDefaultFS)
	//		A dialog is here made available for that purpose
	//
	//		In other applications where FLCs have to be optimised
	//		one may want to delete a loaded FS and load another predefined one
	//		This is why it may be interesting to create many fuzzy sets initially
	//
	//		One can also for instance use an optimisation technique such
	//		Genetic Algorithms to find the best suited FS for a problem on hand.

	BOOL bFS = FLC->boolFS();
	if (!bFS) 
	{
		AfxMessageBox("No Fuzzy Set in this sFLC!");	
		return;
	}

	CFuzzySet *FS = FLC->FSArray();
	static CFuzzySetDesign s_FSdlg;

	s_FSdlg.m_nNumFS = FLC->GetNumFuzzySets();
	s_FSdlg.FS = FS;			// Pass Fuzzy Set array (i.e. pointer) to dialog
	s_FSdlg.m_nSelectedDefaultFSModel = m_nDefaultFuzzySetModel;
	INT_PTR dlgret = s_FSdlg.DoModal();
	if (dlgret == FALSE)
	{
		AddInfoLine("FS Editor could not be displayed.");
		AddInfoLine("Check that 2D NTGraph library has been installed");
		AddInfoLine("Note: OCX must be registered");
		return;
	}
	AddInfoLine("If a FS has been changed, it must be re-assigned to the variable using it");
}

void CFuzzyDlg::ModifyVS() //	Modify Var Set
{
	BOOL bVS = FLC->boolVS();
	if (!bVS) 
	{
		AfxMessageBox("No Variable Set!");	
		return;
	}

	BOOL bFS = FLC->boolFS();
	if (!bFS) 
	{
		AfxMessageBox("No Fuzzy Set!");	
		return;
	}

	CFLVariable *VS = FLC->VarArray();
	CFuzzySet *FS = FLC->FSArray(); 

	static CFuzzyVarDesign s_VSdlg;
	s_VSdlg.nNumVars = FLC->GetNumVars();	
	s_VSdlg.nNumFS = FLC->GetNumFuzzySets();
	s_VSdlg.VS = VS;
	s_VSdlg.FS = FS;
	s_VSdlg.DoModal();

	for (int i=0; i<FLC->GetNumVars(); i++)
	{
		if (!VS[i].HasFS())
		{
			FLC->ClearVarSet();
			AddInfoLine("Variable Set had to be cleared (variable had no FS)");
			return;
		}
	}
}

void CFuzzyDlg::ModifyRS() // Modify Rule Set
{
	if (!FLC->boolRS()) 
	{
		AddInfoLine("No Rule Set. Can't continue...");
		return;
	}

	if (!FLC->boolVS()) 
	{
		AddInfoLine("No Variable Set. Can't continue...");
		return;
	}

	CFLRule *RS = FLC->RuleArray();
	CFLVariable *VS = FLC->VarArray();

	static CRuleMatrix s_RSdlg;
	s_RSdlg.nNumRules = FLC->GetRuleSetSize();
	s_RSdlg.nNumVars = FLC->GetNumVars(); // should be 2+1 anyway
	s_RSdlg.RS = RS;
	s_RSdlg.VS = VS;
	s_RSdlg.DoModal();
}

void CFuzzyDlg::ResetFLC()
{
	FLC->ClearFuzzySets();
	FLC->ClearVarSet();
	FLC->ClearRuleSet();
	AddInfoLine("FLC has been cleared");
}

void CFuzzyDlg::DeleteFS()
{
	FLC->ClearFuzzySets();
	AddInfoLine("Fuzzy Sets have been deleted");
}

void CFuzzyDlg::DeleteVS()
{
	FLC->ClearVarSet();
	AddInfoLine("Variable Set has been deleted");
	if (FLC->boolRS())
	{
		FLC->ClearRuleSet();
		AddInfoLine("Rule Set has also been deleted");
	}
}

void CFuzzyDlg::DeleteRS()	// Delete RS
{
	FLC->ClearRuleSet();
	AddInfoLine("Rule Set has been deleted");
}

void CFuzzyDlg::OnDestroy()
{
	CDialog::OnDestroy();
	FLC->Reset();		// Clear FLC if user presses the X instead of "Close"
}

void CFuzzyDlg::ClearInfoList()
{
	m_Info.ResetContent();
}

void CFuzzyDlg::ClearFLCDisplayList()	// Clear display lists
{
	m_Info.ResetContent();
	lstPredicate.ResetContent();
	lstEvalPredicates.ResetContent();
	lstOutputStrength.ResetContent();
}

void CFuzzyDlg::AddInfoLine(CString str)
{
	m_Info.AddString(str);
	m_Info.SetCurSel(m_Info.GetCount()-1);
}

void CFuzzyDlg::Click_3x3_FS_model() // 3x3 format (3 fuzzy states)
{
	CString str;
	CRect lprect;

	m_Info.ResetContent();

	m_nFLC_Shape = 0;				// Shape = 0 for 3x3
	FLC->SetShape(m_nFLC_Shape);	
	// NumFS3: Number of 3-state Fuzzy Sets to display in listbox
	if (FLC->boolFS())	ResetFLC();

	// Create cbbFS and populate it with available FS models
	cbbFS.ResetContent();
	cbbFS.GetWindowRect( &lprect ); 
	lprect.bottom = lprect.top + NumFS3 * cbbFS.GetItemHeight( -1 ) + lprect.Height();
	cbbFS.SetWindowPos(NULL, 0, 0, 
		lprect.Width(), lprect.Height(), 
		SWP_NOMOVE | SWP_NOZORDER );

	// Display the available Fuzzy Set models
	for (int i=0; i < NumFS3; i++) 
		cbbFS.InsertString(i,FS3n[i].c_str());
	cbbFS.SetCurSel(0);
	m_nDefaultFuzzySetModel = 0;

	// Same for RuleSets
	cbbRS.ResetContent();
	cbbRS.GetWindowRect( &lprect ); 
	lprect.bottom = lprect.top + NumRS33 * cbbRS.GetItemHeight( -1 ) + lprect.Height();
	cbbRS.SetWindowPos(NULL, 0, 0, 
		lprect.Width(), lprect.Height(), 
		SWP_NOMOVE | SWP_NOZORDER );

	// Display the available Rule Set models
	for (int i=0; i < NumRS33; i++) 
		cbbRS.InsertString(i,RS33n[i].c_str());
	cbbRS.SetCurSel(0);
	m_nDefaultRuleSetModel = 0;
	FLC->SetRSModel(RS33n[0]);
	AddInfoLine("sFLC Model is 3x3: 3 fuzzy states par variable, and 3x3 Rule Set");
}

void CFuzzyDlg::Click_5x5_FS_model() // 5x5 format (5 fuzzy states)
{
	CString str;
	CRect lprect;

	m_Info.ResetContent();

	m_nFLC_Shape = 1;				// Shape = 0 for 3x3
	FLC->SetShape(1);	// Shape = 1 for 5x5

	// NumFS5: Number of 5x5 Fuzzy Sets to display in listbox

	if (FLC->boolFS())	ResetFLC();

	cbbFS.ResetContent();
	cbbFS.GetWindowRect( &lprect ); 
	lprect.bottom = lprect.top + NumFS5 * cbbFS.GetItemHeight( -1 ) + lprect.Height();
	cbbFS.SetWindowPos(NULL, 0, 0, 
		lprect.Width(), lprect.Height(), 
		SWP_NOMOVE | SWP_NOZORDER );

	// Display the available Fuzzy Set Models
	for (int i=0; i < NumFS5; i++) 
		cbbFS.InsertString(i,FS5n[i].c_str());
	cbbFS.SetCurSel(0);
	m_nDefaultFuzzySetModel = 0;

	cbbRS.ResetContent();
	cbbRS.GetWindowRect( &lprect ); 
	lprect.bottom = lprect.top + NumRS55 * cbbRS.GetItemHeight( -1 ) + lprect.Height();
	cbbRS.SetWindowPos(NULL, 0, 0, 
		lprect.Width(), lprect.Height(), 
		SWP_NOMOVE | SWP_NOZORDER );

	// Display the available Rule Set Modelss
	for (int i=0; i < NumRS55; i++) 
		cbbRS.InsertString(i,RS55n[i].c_str());
	cbbRS.SetCurSel(0);
	m_nDefaultRuleSetModel = 0;
	FLC->SetRSModel(RS55n[0]);
	AddInfoLine("sFLC Model is 5x5: 5 fuzzy states par variable, and 5x5 Rule Set");
}

void CFuzzyDlg::Select_FSmodel()
{	// from Fuzzy Sets combobox
	m_nDefaultFuzzySetModel = cbbFS.GetCurSel();
}

void CFuzzyDlg::Select_RSmodel()	// Default Rule Set Model
{
	m_nDefaultRuleSetModel = cbbRS.GetCurSel();
	if (FLC->GetShape() == 0)
		FLC->SetRSModel(RS33n[m_nDefaultRuleSetModel]);
	else
		FLC->SetRSModel(RS55n[m_nDefaultRuleSetModel]);
}

void CFuzzyDlg::OnBnClickedShape()
{
	if (m_nFLC_Shape < 0) return;

	FS_shape SD;	// Dialog showing FS shapes
	SD.m_nShape = m_nFLC_Shape;
	SD.m_nFSnum = m_nDefaultFuzzySetModel;
	SD.DoModal();
}

void CFuzzyDlg::OnBnClickedWeb()
{
CString strURL= "http://www.foretrade.com/sflc3.htm";
	::ShellExecute(NULL, NULL, strURL, NULL, NULL, SW_SHOWNORMAL);
}
