// FuzzySetDesign.cpp : implementation file
//
#include "stdafx.h"
#include "Fuzzy.h"
#include "FuzzySetDesign.h"

#include "NewCellTypes/GridURLCell.h"
#include "NewCellTypes/GridCellCombo.h"
#include "NewCellTypes/GridCellCheck.h"
#include "NewCellTypes/GridCellNumeric.h"
#include "NewCellTypes/GridCellDateTime.h"



BEGIN_EVENTSINK_MAP(CFuzzySetDesign, CDialog)
	ON_EVENT(CFuzzySetDesign, IDC_NTGRAPHCTRL1, DISPID_CLICK, ClickNtgraphctrl1, VTS_NONE)
END_EVENTSINK_MAP()

// CFuzzySetDesign dialog

IMPLEMENT_DYNAMIC(CFuzzySetDesign, CDialog)
CFuzzySetDesign::CFuzzySetDesign(CWnd* pParent /*=NULL*/)
	: CDialog(CFuzzySetDesign::IDD, pParent)
	, m_nNumFS(-1)
	, m_nSelectedFS(-1)
	, m_nSelectedDefaultFSModel(-1)
{	// Each FS has its model and its labels
}

CFuzzySetDesign::~CFuzzySetDesign()
{
}

void CFuzzySetDesign::PostNcDestroy()
{
	delete [] m_nFSwording;	// 'Fuzzy States labels'
	delete [] m_nFSmodel;	// Fuzzy model
	delete [] m_bFSmodelmod;
	CDialog::PostNcDestroy();
}


void CFuzzySetDesign::DoDataExchange(CDataExchange* pDX)
{
	CDialog::DoDataExchange(pDX);
	DDX_GridControl(pDX, IDC_FS_GRID, m_Grid);
	DDX_Control(pDX, IDC_NTGRAPHCTRL1, m_Graph);
	DDX_Control(pDX, IDC_COMBO1, m_FStatesNames);
	DDX_Control(pDX, IDC_COMBO2, lstFS);
	DDX_Control(pDX, IDC_LIST_NUMSTATES, lstNumStates);
	DDX_Control(pDX, IDC_COMBO3, lstModel);
	DDX_Control(pDX, IDC_EDIT1, edtFSModel);
}

BEGIN_MESSAGE_MAP(CFuzzySetDesign, CDialog)
	ON_CBN_SELCHANGE(IDC_COMBO1, ChangeFStatesLabels)
	ON_BN_CLICKED(ID_DRAW, DrawMF)
	ON_CBN_SELCHANGE(IDC_COMBO2, SelectFuzzySet)
	ON_BN_CLICKED(IDC_BUTTON1, SaveChanges)
	ON_BN_CLICKED(IDOK, CloseDialog)
	ON_CBN_SELCHANGE(IDC_COMBO3, &CFuzzySetDesign::ChangeFuzzyModel)
	ON_LBN_SELCHANGE(IDC_LIST_NUMSTATES, &CFuzzySetDesign::OnChangeNumstates)
END_MESSAGE_MAP()

// CFuzzySetDesign message handlers

BOOL CFuzzySetDesign::OnInitDialog()
{
	CDialog::OnInitDialog();

	m_nFSwording = new int[m_nNumFS];
	m_nFSmodel = new int[m_nNumFS];
	m_bFSmodelmod = new BOOL[m_nNumFS];

	// We replicate the default FS model to all FS for now
	for (int i = 0 ; i < m_nNumFS; i++)
		m_nFSmodel[i] = FS[i].GetFSmodel();

	CString str;
	CRect lprect;

	int nLines = m_nNumFS;	// Num Fuzzy Sets to display in listbox
	str.Format("Num Fuzzy Sets in this sFLC: %i",m_nNumFS);
	SetWindowText(str);		// Change window caption

	lstFS.GetWindowRect( &lprect ); 
	lprect.bottom = lprect.top + nLines * lstFS.GetItemHeight( -1 ) + lprect.Height();
	lstFS.SetWindowPos(NULL, 0, 0, 
		lprect.Width(), lprect.Height(), 
		SWP_NOMOVE | SWP_NOZORDER );

	// Display name of the available Fuzzy Sets
	for (int i=0; i < m_nNumFS; i++) 
		lstFS.AddString(FS[i].GetFSName().c_str());
	lstFS.SetCurSel(0);

	// some initial states
	int nW = FS[0].GetFSwording();		//  To select correct state labels.
	int nNumStates = FS[0].NumStates(); //	equivalent to model shape

	lstNumStates.AddString("3");
	lstNumStates.AddString("5");
	lstNumStates.AddString("7");

	nLines = (nNumStates == 3 ? NumFS3 : NumFS5);  // temp solution
	lstModel.GetWindowRect( &lprect ); 
	lprect.bottom = lprect.top + nLines * lstModel.GetItemHeight( -1 ) + lprect.Height();
	lstModel.SetWindowPos(NULL, 0, 0, 
		lprect.Width(), lprect.Height(), 
		SWP_NOMOVE | SWP_NOZORDER );

	switch (nNumStates)
	{	// we suppose all fuzzy set have the same number of states
		// we now populate lstNumStates with existing predefined models
		default:
		case 3:
			for (int i = 0; i < NumFS3; i++)
				lstModel.InsertString(i,FS3n[i].c_str());
			lstNumStates.SetCurSel(0);
			edtFSModel.SetWindowTextA(FS3n[m_nFSmodel[0]].c_str());
			m_nRows = 3;
			break;
		case 5:
			for (int i = 0; i < NumFS5; i++)
				lstModel.InsertString(i,FS5n[i].c_str());
			edtFSModel.SetWindowTextA(FS5n[m_nFSmodel[0]].c_str());
			lstNumStates.SetCurSel(1);
			m_nRows = 5;
			break;
		case 7:
			for (int i = 0; i < NumFS7; i++)
				lstModel.InsertString(i,FS7n[i].c_str());
			lstNumStates.SetCurSel(2);
			m_nRows = 7;
			break;
	}

	lstModel.SetCurSel(m_nFSmodel[0]); // initially

	// Set up Grid & Graph
	m_Grid.DeleteAllItems();
	m_Grid.EnableDragAndDrop(FALSE);
	m_Grid.GetDefaultCell(FALSE, FALSE)->SetBackClr(RGB(0xFF, 0xFF, 0xE0));

	m_Grid.EnableTitleTips(FALSE);

	m_Grid.SetFixedColumnSelection(TRUE);
    m_Grid.SetFixedRowSelection(TRUE);
    m_Grid.AutoSize();
	m_Grid.SetCompareFunction(CGridCtrl::pfnCellNumericCompare);

	m_nFixCols = 1;
	m_nFixRows = 1;

	m_Grid.SetGridLines(GVL_BOTH);
	m_Grid.SetAutoSizeStyle();
    
	m_nCols = 8;

	m_Grid.SetRowCount(m_nRows);
	m_Grid.SetColumnCount(m_nCols);
	m_Grid.SetFixedRowCount(m_nFixRows);
	m_Grid.SetFixedColumnCount(m_nFixCols);

	m_Grid.SetEditable(TRUE);
    
	m_Grid.SetItemText(0,1,"Name");
	m_Grid.SetItemText(0,2,"Type");
	m_Grid.SetItemText(0,3,"Point1");
	m_Grid.SetItemText(0,4,"Point2");
	m_Grid.SetItemText(0,5,"Point3");
	m_Grid.SetItemText(0,6,"Point4");
	m_Grid.SetItemText(0,7,"Centre");

	CRect Rect;
	m_Grid.GetClientRect(&Rect);
	m_Grid.SetColumnWidth(0,35);
	for (int i=1;i<m_Grid.GetColumnCount();i++)
		m_Grid.SetColumnWidth(i,(Rect.Width()-35)/7);

	// Graph
	m_Graph.SetLabelColor (RGB(128,128,255));
	m_Graph.SetPlotAreaColor(RGB(150,150,150));
	m_Graph.SetFrameStyle(1);	// (1) - Flat // (2) - Scope (raised frame and sunken plot area borders)
								// (3) - 3DFrame (a bitmap frame picture)
    
	m_Graph.SetGridColor(RGB(255,192,192));
	m_Graph.SetShowGrid (TRUE);
	
	m_Graph.SetCursorColor (RGB(255,0,0));
	m_Graph.SetTrackMode (1);

	m_Graph.SetCaption("Fuzzy Set");
	m_Graph.SetXLabel("Value");
	m_Graph.SetYLabel("DOM");
	m_Graph.SetRange(-1.05,1.05,-0.05,1.05);
	// End Grid & Graph

	m_FStatesNames.ResetContent();

	SelectFuzzySet();			// Select Fuzzy Set
	ChangeFuzzyModel();			// Change the different fuzzy models;
	ChangeFStatesLabels();		// Change names
	return TRUE;
}

void CFuzzySetDesign::ChangeFStatesLabels() // Change F.States names
{	// Apply new labels to grid, then save
	int n = m_FStatesNames.GetCurSel();
	for (int i = 1; i <= m_nRows; i++)
	{
		switch (m_nRows)
		{
		default:
		case 3:
			m_Grid.SetItemText(i,1,strStates3[n][i-1].c_str());
			break;
		case 5:
			m_Grid.SetItemText(i,1,strStates5[n][i-1].c_str());
			break;
		case 7:
			m_Grid.SetItemText(i,1,strStates7[n][i-1].c_str());
			break;
		}
	}
	SaveChanges();	// save changes to MF settings
	m_Grid.Refresh();
}

void CFuzzySetDesign::ClickNtgraphctrl1()
{
	m_Graph.ShowProperties();
}

void CFuzzySetDesign::DrawMF() // Draw MF
{
	int nPoints = 201;
	int nElements = m_nRows;

	m_Graph.ClearGraph();
	if (nElements <= 1) return;
	m_Graph.SetElementWidth(2);

	//int n = m_Graph.GetElementCount();
	for (int i = 1; i < m_nRows; i++)
	{
		m_Graph.AddElement();
		m_Graph.SetElementLineColor (RGB(0,255-50*i,40*i));
		m_Graph.SetElementWidth(2);
		m_Graph.SetElementLinetype(1);
	}

	int n = m_Graph.GetElementCount();
	m_Graph.SetRange(-1.1, 1.1, -0.1, 1.1);

	for (short i = 0; i < n; i++) 
	{
		CString str = m_Grid.GetItemText(i+1,2);
		int nType = atol(str);
		str = m_Grid.GetItemText(i+1,3);
		double dPoint1 = atof(str);
		str = m_Grid.GetItemText(i+1,4);
		double dPoint2 = atof(str);

		for (int x = -100; x <= 100; x++) 
		{
			double y;
			double X = x / 100;
			switch (nType)
			{
				case 0:
					y = mfShoulderRight(X,dPoint1,dPoint2);
					break;
				case 1:
					y = mfTriangle(X,dPoint1,dPoint2);
					break;
				case 2:
					y = mfShoulderLeft(X,dPoint1,dPoint2);
					break;
				default:
					y = (double)rand()/RAND_MAX;		// goes loony
					break;
			}
			m_Graph.PlotXY(X,y,i);
		}
	}
}

// Extract end displau FS Model details
void CFuzzySetDesign::SelectFuzzySet()
{
	m_nSelectedFS = lstFS.GetCurSel();				//  m_nSelectedFS keeps track of the current FS
	m_nRows = FS[m_nSelectedFS].NumStates();		//  1 row per fuzzy state on the grid
	m_nFSwording[m_nSelectedFS] = FS[m_nSelectedFS].GetFSwording();
	m_nFSmodel[m_nSelectedFS] = FS[m_nSelectedFS].GetFSmodel();
	string model;
	
	switch (FS[m_nSelectedFS].NumStates())
	{
	default:
	case 3:
		model.assign(FS3n[m_nFSmodel[m_nSelectedFS]]);
		break;
	case 5:
		model.assign(FS5n[m_nFSmodel[m_nSelectedFS]]);
		break;
	case 7:
		model.assign(FS7n[m_nFSmodel[m_nSelectedFS]]);
		break;
	}
	edtFSModel.SetWindowTextA(model.c_str());
//	lstModel.SetCurSel(FS[m_nSelectedFS].GetFSmodel());
	m_Grid.DeleteNonFixedRows();
	m_Grid.Refresh();
	m_Graph.ClearGraph(); 

	ChangeFuzzyModel();
	// Can't draw right now, hence press Draw button to redraw
}

void CFuzzySetDesign::SaveChanges()	// save changes to MF settings
{	// FuzzySet Name can't be changed here
	CFLMembershipFunction *MF = FS[m_nSelectedFS].FuzzyStateArray();

	FS[m_nSelectedFS].SetFSwording(m_FStatesNames.GetCurSel());

	for (int i = 1; i <= m_nRows; i++)
	{	// save shape
		_MFshape mf;
		CString str;
		
		str = m_Grid.GetItemText(i,1);
		mf.strName = str;

		str = m_Grid.GetItemText(i,2);
		mf.nType = atol(str);

		str = m_Grid.GetItemText(i,3);
		mf.dPoint1 = atof(str);

		str = m_Grid.GetItemText(i,4);
		mf.dPoint2 = atof(str);

		str = m_Grid.GetItemText(i,5);
		mf.dPoint3 = atof(str);

		str = m_Grid.GetItemText(i,6);
		mf.dPoint4 = atof(str);

		str = m_Grid.GetItemText(i,7);
		mf.dCenter  = atof(str);

		MF[i-1].SetMFshape(mf);
	}
}
void CFuzzySetDesign::CloseDialog()
{
	OnOK();
}


void CFuzzySetDesign::ChangeFuzzyModel()
{
	//	m_nFSmodel = lstModel.GetCurSel();

	const CString s = " - ";
	m_FStatesNames.ResetContent(); // list of attributes (names)

	CRect lprect;
	int nLines = m_nRows;
	m_FStatesNames.GetWindowRect( &lprect ); 
	lprect.bottom = lprect.top + nLines * m_FStatesNames.GetItemHeight( -1 ) + lprect.Height();
	m_FStatesNames.SetWindowPos(NULL, 0, 0, lprect.Width(), lprect.Height(), SWP_NOMOVE | SWP_NOZORDER );

	// set 1st element
	m_Graph.SetElementLineColor(RGB(255,0,0));
	m_Graph.SetElementLinetype(0);
	m_Graph.SetElementWidth(1);

	m_Graph.SetElementPointColor(RGB(0,0,255));
	m_Graph.SetElementPointSymbol(3);
	m_Graph.SetElementSolidPoint(TRUE);

	// Can't change NumStates here
	switch (FS[m_nSelectedFS].NumStates())
	{
	case 3:
	{
		m_nRows = 3;
		m_Grid.SetRowCount(m_nRows);
		// Fill listbox with fuzzy attributes (3)
		m_FStatesNames.InsertString(0,
			strStates3[0][0].c_str()+s+strStates3[0][1].c_str()+s+strStates3[0][2].c_str());
		m_FStatesNames.InsertString(1,
			strStates3[1][0].c_str()+s+strStates3[1][1].c_str()+s+strStates3[1][2].c_str());
		m_FStatesNames.InsertString(2,
			strStates3[2][0].c_str()+s+strStates3[2][1].c_str()+s+strStates3[2][2].c_str());
		m_FStatesNames.SetCurSel(m_nFSwording[m_nSelectedFS]);

	    m_Grid.SetRowCount(m_nRows+1);
		for (int i = 1; i <= m_nRows; i++)
		{
			_MFshape mf = FS[m_nSelectedFS].FuzzyStateElt(i-1).MFshape();
			CString str;
			str.Format("MF%i",i);
			m_Grid.SetItemText(i,0,str);
			if (mf.strName == "")
				m_Grid.SetItemText(i,1,strStates3[0][i-1].c_str());
			else
				m_Grid.SetItemText(i,1,mf.strName.c_str());
			m_Grid.SetCellType(i,2,RUNTIME_CLASS(CGridCellNumeric));	// MFShape type

			str.Format("%i",mf.nType);
			m_Grid.SetItemText(i,2,str);
			str.Format("%.2f",mf.dPoint1);
			m_Grid.SetItemText(i,3,str);
			str.Format("%.2f",mf.dPoint2);
			m_Grid.SetItemText(i,4,str);
			str.Format("%.2f",mf.dPoint3);
			m_Grid.SetItemText(i,5,str);
			str.Format("%.2f",mf.dPoint4);
			m_Grid.SetItemText(i,6,str);
			str.Format("%.2f",mf.dCenter);
			m_Grid.SetItemText(i,7,str);
		}
		lstNumStates.SetCurSel(m_nFSmodel[m_nSelectedFS]);
		break;
	}
	case 5:
	{
		m_nRows = 5;
		m_Grid.SetRowCount(m_nRows);
		// Fill listbox with fuzzy attributes (5)
		m_FStatesNames.InsertString(0,
				strStates5[0][0].c_str()+s+strStates5[0][1].c_str()+s+
				strStates5[0][2].c_str()+s+strStates5[0][3].c_str()+s+
				strStates5[0][4].c_str());
		m_FStatesNames.InsertString(1,
				strStates5[1][0].c_str()+s+strStates5[1][1].c_str()+s+
				strStates5[1][2].c_str()+s+strStates5[1][3].c_str()+s+
				strStates5[1][4].c_str());
		m_FStatesNames.InsertString(2,
				strStates5[2][0].c_str()+s+strStates5[2][1].c_str()+s+
				strStates5[2][2].c_str()+s+strStates5[2][3].c_str()+s+
				strStates5[2][4].c_str());
		m_FStatesNames.InsertString(3,
				strStates5[3][0].c_str()+s+strStates5[3][1].c_str()+s+
				strStates5[3][2].c_str()+s+strStates5[3][3].c_str()+s+
				strStates5[3][4].c_str());
		m_FStatesNames.InsertString(4,
				strStates5[4][0].c_str()+s+strStates5[4][1].c_str()+s+
				strStates5[4][2].c_str()+s+strStates5[4][3].c_str()+s+
				strStates5[4][4].c_str());

		m_FStatesNames.SetCurSel(0);

		m_Grid.SetRowCount(m_nRows+1);
		for (int i=1;i<=m_nRows;i++)
		{
			_MFshape mf = FS[m_nSelectedFS].FuzzyStateElt(i-1).MFshape();
			CString str;
			str.Format("MF%i",i);
			m_Grid.SetItemText(i,0,str);
			m_Grid.SetItemText(i,1,strStates5[0][i-1].c_str());
			m_Grid.SetCellType(i,2,RUNTIME_CLASS(CGridCellNumeric));	// MFShape type
					
			str.Format("%i",mf.nType);
			m_Grid.SetItemText(i,2,str);
			str.Format("%.2f",mf.dPoint1);
			m_Grid.SetItemText(i,3,str);
			str.Format("%.2f",mf.dPoint2);
			m_Grid.SetItemText(i,4,str);
			str.Format("%.2f",mf.dPoint3);
			m_Grid.SetItemText(i,5,str);
			str.Format("%.2f",mf.dPoint4);
			m_Grid.SetItemText(i,6,str);
			str.Format("%.2f",mf.dCenter);
			m_Grid.SetItemText(i,7,str);
		}
		break;
	}
	case 7:
	{
		m_nRows = 7;
		m_Grid.SetRowCount(m_nRows);
		// Fill listbox with fuzzy attributes (5)
		m_FStatesNames.InsertString(0,
					strStates7[0][0].c_str()+s+strStates7[0][1].c_str()+s+
					strStates7[0][2].c_str()+s+strStates7[0][3].c_str()+s+
					strStates7[0][4].c_str()+s+strStates7[0][5].c_str()+s+
					strStates7[0][6].c_str());
		m_FStatesNames.InsertString(1,
					strStates7[1][0].c_str()+s+strStates7[1][1].c_str()+s+
					strStates7[1][2].c_str()+s+strStates7[1][3].c_str()+s+
					strStates7[1][4].c_str()+s+strStates7[1][5].c_str()+s+
					strStates7[1][6].c_str());
		m_FStatesNames.InsertString(2,
					strStates7[2][0].c_str()+s+strStates7[2][1].c_str()+s+
					strStates7[2][2].c_str()+s+strStates7[2][3].c_str()+s+
					strStates7[2][4].c_str()+s+strStates7[2][5].c_str()+s+
					strStates7[2][6].c_str());
		m_FStatesNames.InsertString(3,
					strStates7[3][0].c_str()+s+strStates7[3][1].c_str()+s+
					strStates7[3][2].c_str()+s+strStates7[3][3].c_str()+s+
					strStates7[3][4].c_str()+s+strStates7[3][5].c_str()+s+
					strStates7[3][6].c_str());
		m_FStatesNames.InsertString(4,
					strStates7[4][0].c_str()+s+strStates7[4][1].c_str()+s+
					strStates7[4][2].c_str()+s+strStates7[4][3].c_str()+s+
					strStates7[4][4].c_str()+s+strStates7[4][5].c_str()+s+
					strStates7[4][6].c_str());
		m_FStatesNames.InsertString(5,
					strStates7[5][0].c_str()+s+strStates7[5][1].c_str()+s+
					strStates7[5][2].c_str()+s+strStates7[5][3].c_str()+s+
					strStates7[5][4].c_str()+s+strStates7[5][5].c_str()+s+
					strStates7[5][6].c_str());
		m_FStatesNames.InsertString(6,
					strStates7[6][0].c_str()+s+strStates7[6][1].c_str()+s+
					strStates7[6][2].c_str()+s+strStates7[6][3].c_str()+s+
					strStates7[6][4].c_str()+s+strStates7[6][5].c_str()+s+
					strStates7[6][6].c_str());

		m_Grid.SetRowCount(m_nRows+1);
		for (int i=1;i<=m_nRows;i++)
		{
			_MFshape mf = FS[m_nSelectedFS].FuzzyStateElt(i-1).MFshape();
			CString str;
			str.Format("MF%i",i);
			m_Grid.SetItemText(i,0,str);
			m_Grid.SetItemText(i,1,strStates5[0][i-1].c_str());
			m_Grid.SetCellType(i,2,RUNTIME_CLASS(CGridCellNumeric));	// MFShape type
					
			str.Format("%i",mf.nType);
			m_Grid.SetItemText(i,2,str);
			str.Format("%.2f",mf.dPoint1);
			m_Grid.SetItemText(i,3,str);
			str.Format("%.2f",mf.dPoint2);
			m_Grid.SetItemText(i,4,str);
			str.Format("%.2f",mf.dPoint3);
			m_Grid.SetItemText(i,5,str);
			str.Format("%.2f",mf.dPoint4);
			m_Grid.SetItemText(i,6,str);
			str.Format("%.2f",mf.dCenter);
			m_Grid.SetItemText(i,7,str);
		}
	}
	}
	lstNumStates.SetCurSel(m_nFSmodel[m_nSelectedFS]);
}

void CFuzzySetDesign::OnChangeNumstates()
{
	AfxMessageBox("Sorry, model size cannot be changed here");
}
