// FuzzyVarDesign.cpp : implementation file
//

#include "stdafx.h"
#include "Fuzzy.h"
#include "FuzzyVarDesign.h"

#include "NewCellTypes/GridURLCell.h"
#include "NewCellTypes/GridCellCombo.h"
#include "NewCellTypes/GridCellCheck.h"
#include "NewCellTypes/GridCellNumeric.h"
#include "NewCellTypes/GridCellDateTime.h"

// CFuzzyVarDesign dialog

IMPLEMENT_DYNAMIC(CFuzzyVarDesign, CDialog)
CFuzzyVarDesign::CFuzzyVarDesign(CWnd* pParent /*=NULL*/)
	: CDialog(CFuzzyVarDesign::IDD, pParent)
		, nNumVars(0)		// Num Variables
		, nNumFS(0)			// Num Fuzzy Sets
{
}

CFuzzyVarDesign::~CFuzzyVarDesign()
{
}

void CFuzzyVarDesign::DoDataExchange(CDataExchange* pDX)
{
	CDialog::DoDataExchange(pDX);
	DDX_GridControl(pDX, IDC_VARFUZZ, m_Grid);	// display grid
}


BEGIN_MESSAGE_MAP(CFuzzyVarDesign, CDialog)
	ON_BN_CLICKED(IDOK, OnBnClickedOk)
END_MESSAGE_MAP()


// CFuzzyVarDesign message handlers

BOOL CFuzzyVarDesign::OnInitDialog()
{
	CDialog::OnInitDialog();

	CString str;
	CRect lprect;

	str.Format("Variables : %i",nNumVars);
	SetWindowText(str);		// Change window caption

	m_nRows = nNumVars;

	m_Grid.DeleteAllItems();
	m_Grid.EnableDragAndDrop(FALSE);
	m_Grid.GetDefaultCell(FALSE, FALSE)->SetBackClr(RGB(0xFF, 0xFF, 0xE0));

	m_Grid.SetFixedColumnSelection(TRUE);
    m_Grid.SetFixedRowSelection(TRUE);
    m_Grid.AutoSize();
	m_Grid.SetCompareFunction(CGridCtrl::pfnCellNumericCompare);

	m_nFixCols = 1;
	m_nFixRows = 1;

	m_Grid.SetGridLines(GVL_BOTH);
	m_Grid.SetAutoSizeStyle();

	m_nRows = nNumVars;
	m_nCols = 6;

	TRY {
	    m_Grid.SetFixedRowCount(m_nFixRows);
	    m_Grid.SetFixedColumnCount(m_nFixCols);
		m_Grid.SetRowCount(m_nRows+1);
	    m_Grid.SetColumnCount(m_nCols);
	}
	CATCH (CMemoryException, e)
	{
	   	e->ReportError();
    	return FALSE;
	}
    END_CATCH

	m_Grid.SetEditable(TRUE);
    
	m_Grid.SetItemText(0,1,"VarNo");
	m_Grid.SetItemText(0,2,"Name");
	m_Grid.SetItemText(0,3,"Input");
	m_Grid.SetItemText(0,4,"States");
	m_Grid.SetItemText(0,5,"FuzzySet");
	//m_Grid.SetItemText(0,6,"FSno");

    CStringArray VarIO;			// to display a dropdown control on the grid
    VarIO.Add(_T("Input"));
    VarIO.Add(_T("Output"));

    CStringArray VarFS;			// to display a dropdown control on the grid
	VarFS.Add("None");
	for (int i=0; i<nNumFS; i++)// each FS name is added to the string array
		VarFS.Add(FS[i].GetFSName().c_str());

	CRect Rect;
	m_Grid.GetClientRect(&Rect);
	
	m_Grid.SetColumnWidth(0,20);
	m_Grid.SetColumnWidth(1,20);
	int w = 30;
	for (int i=2;i<m_Grid.GetColumnCount()-1;i++)
	{
		int cw = (Rect.Width()-50)/(m_nCols-1);
		m_Grid.SetColumnWidth(i,cw);
		w += cw;
	}
	m_Grid.SetColumnWidth(m_Grid.GetColumnCount()-1,Rect.Width()- w - 10);

	for (int i=0;i<m_nRows;i++)
	{
		CString str;
	    m_Grid.SetItemState(i+1,1, m_Grid.GetItemState(i+1,1) | GVIS_READONLY);
		m_Grid.SetCellType(i+1,1,RUNTIME_CLASS(CGridCellNumeric));
		str.Format(_T("%i"),VS[i].VarNumber() );				// display Var Number
		m_Grid.SetItemText(i+1,1,str);
		m_Grid.SetItemText(i+1,2,VS[i].Name().c_str());		// display Var Name

		m_Grid.SetCellType(i+1,3, RUNTIME_CLASS(CGridCellCombo));
	    CGridCellCombo *pCell1 = (CGridCellCombo*) m_Grid.GetCell(i+1,3);
		pCell1->SetOptions(VarIO);							// display Input/Output dropdown
		pCell1->SetStyle(CBS_DROPDOWN); //CBS_DROPDOWN, CBS_DROPDOWNLIST, CBS_SIMPLE

		if (VS[i].IsInput()) 
			m_Grid.SetItemText(i+1,3, _T("Input"));			// display Input or Output
		else
			m_Grid.SetItemText(i+1,3, _T("Output"));

	    m_Grid.SetItemState(i+1,4, m_Grid.GetItemState(i+1,4) | GVIS_READONLY);
		str.Format("%i",VS[i].NumStates());					// display num fuzzy states
		m_Grid.SetItemText(i+1,4,str);

		m_Grid.SetCellType(i+1,5, RUNTIME_CLASS(CGridCellCombo));
	    CGridCellCombo *pCell2 = (CGridCellCombo*) m_Grid.GetCell(i+1,5);
		pCell2->SetOptions(VarFS);			// Fuzzy Set names
		pCell2->SetStyle(CBS_DROPDOWN); //CBS_DROPDOWN, CBS_DROPDOWNLIST, CBS_SIMPLE
		
		if (!VS[i].HasFS()) 
			m_Grid.SetItemText(i+1,5,"None");
		else								// the correct FS name is displayed
			m_Grid.SetItemText(i+1,5,VS[i].FuzzySet()->GetFSName().c_str());

		//str.Format("%i",VS[i].FuzzySet()->FSNumber());
		//m_Grid.SetItemText(i+1,6,str);
	}
	return TRUE;  // return TRUE unless you set the focus to a control
}

void CFuzzyVarDesign::OnBnClickedOk()
{
	for (int i=0;i<m_nRows;i++)	// for each variable
	{
		VS[i].SetName(m_Grid.GetItemText(i+1,2).GetString());
		BOOL bInput = (m_Grid.GetItemText(i+1,3) == "Input");
        VS[i].SetInput(bInput);
		if (m_Grid.GetItemText(i+1,5) == "None")
		{
			VS[i].UnloadFuzzySet();
			VS[i].ResetFSbool();
		}
		else
		{
			BOOL found = FALSE;
			for (int j=0; j<nNumFS; j++)
			{	
				if (m_Grid.GetItemText(i+1,5) == FS[j].GetFSName().c_str())
				{
					found = TRUE;
					VS[i].UnloadFuzzySet();
					VS[i].LoadFuzzySet(FS[j], bInput);
				}
			}
			if (!found)
			{
				AfxMessageBox("Unknown Fuzzy Set.  Update is cancelled");
				return;
			}
		}
	}
	OnOK();
}
